#include "board.h"
#include "io.h"
#include "common.h"
#include "types.h"
#include "sdhi.h"
#include "mmcsd.h"
#include "delay.h"
#include "print.h"

static int get_sdhi_port(u32 base)
{
	int port;

	switch (base) {
	case SDHI0_BASE_ADDR:
		port = 0;
		break;
	case SDHI1_BASE_ADDR:
		port = 1;
		break;
	default:
		port = -1;
		break;
	}

	return port;
}

#define SDHI_READ16(base, addr) \
	read16(base + (addr << (get_sdhi_port(base) ? 0 : 1)))

#define SDHI_WRITE16(base, addr, val) \
	write16(base + (addr << (get_sdhi_port(base) ? 0 : 1)), val)

static void sdhi_pinmux(int port)
{
	u32 reg;

	reg = read32(PFC_IPSR13);
	switch (port) {
	case 0:
	default:
		/* SD0_CLK, SD0_CMD, SD0_DATA[4:0], SD0_CD, SD0_WP */
		reg &= ~0x003ffC00;
		reg |=  0x00000000;
		break;
	case 1:
		/* SD1_CLK, SD1_CMD, SD1_DATA[4:0], SD1_CD */
		reg &= ~0x7fc00000;
		reg |=  0x00000000;
	}
	write32(PFC_PMMR, ~reg);
	write32(PFC_IPSR13, reg);

	if (port == 1) {
		reg = read32(PFC_IPSR14);
		/* SD1_WP */
		reg &= ~0x00000003;
		reg |=  0x00000000;
		write32(PFC_PMMR, ~reg);
		write32(PFC_IPSR14, reg);
	}

	reg = read32(PFC_GPSR6);
	switch (port) {
	case 0:
	default:
		/* GP6[7:0] -> Peripheral(SD0) */
		/* GP6[else] -> GPIO */
		reg |=  0x000000ff;
		break;
	case 1:
		/* GP6[15:8] -> Peripheral(SD1) */
		/* GP6[else] -> GPIO */
		reg |=  0x0000ff00;
	}
	write32(PFC_PMMR, ~reg);
	write32(PFC_GPSR6, reg);
}

static void sdhi_cpg_enable(int port)
{
	u32 sr, cr;
	u32 bit;

	if (port == 1)
		bit = CPG_MSTP3_SDHI1;
	else
		bit = CPG_MSTP3_SDHI0;

	sr = read32(CPG_MSTPSR3);
	if (sr & bit) {
		cr = read32(CPG_SMSTPCR3);
		write32(CPG_SMSTPCR3, cr & ~bit);
	} while (read32(CPG_MSTPSR3) & bit);
}

static void sdhi_clock_start(struct mmc_host *host)
{
	u16 val;

	while (!(SDHI_READ16(host->base, SDHI_INFO2) & 0x2000));
	val = SDHI_READ16(host->base, SDHI_CLK_CTRL);
	val |= 0x0300;
	SDHI_WRITE16(host->base, SDHI_CLK_CTRL, val);

}

static void sdhi_clock_stop(struct mmc_host *host)
{
	u32 val;

	while (!(SDHI_READ16(host->base, SDHI_INFO2) & 0x2000));
	val = SDHI_READ16(host->base, SDHI_CLK_CTRL);
	val &= ~0x0300;
	SDHI_WRITE16(host->base, SDHI_CLK_CTRL, val);
}

static int sdhi_set_clock(struct mmc_host *host, u32 clock)
{
	u32 val;
	int port;

	/* Waiting for SD Bus busy to be cleared */
	while (!(SDHI_READ16(host->base, SDHI_INFO2) & 0x2000));

	val = SDHI_READ16(host->base, SDHI_CLK_CTRL);
	val &= ~0xff;

	port = get_sdhi_port(host->base);
	if (clock) {
		/* high speed: 24.375MHz */
		if (port)
			val |= 0x00;
		else
			val |= 0x01;
	} else {
		/* low speed: 190.4kHz */
		if (port)
			val |= 0x40;
		else
			val |= 0x80;
	}
	SDHI_WRITE16(host->base, SDHI_CLK_CTRL, val);

	return 0;
}

static int sdhi_pio_transfer_data(struct mmc_host *host,
				   struct mmc_command *cmd)
{
	u16 *buffer = (u16 *)cmd->data->buf;
	int i;
	u16 val;

	while(!(SDHI_READ16(host->base, SDHI_INFO2) & 0x0100));
	if (!(SDHI_READ16(host->base, SDHI_INFO2) & 0x0100))
		printf("DATA not Ready\n");

	for (i=0; i<cmd->data->blocks; i++) {
		int blksz = (int)cmd->data->blksz;
		u16 stat;

		do {
			stat = SDHI_READ16(host->base, SDHI_INFO2);
			if (stat & 0x100)
				break;
			if (stat & 0x7f)
				printf("status = %p\n", stat);
		} while(1);

		SDHI_WRITE16(host->base, SDHI_INFO2, 0xfeff);
		val = SDHI_READ16(host->base, SDHI_EXT_ACC);

		if (get_sdhi_port(host->base) == 0)
			SDHI_WRITE16(host->base, SDHI_EXT_ACC, val | (1 << 0));
		else 
			SDHI_WRITE16(host->base, SDHI_EXT_ACC, val & ~(1 << 0));

		while (blksz > 0) {
			*(buffer++) = SDHI_READ16(host->base, SDHI_BUF);
			blksz -= 2;
		}
	}

	return 0;
}

static int sdhi_do_cmd(struct mmc_host *host, struct mmc_command *cmd)
{
	u16 val;
	u32 mode;
	int i;
	u32 stat, stat2;
	static int acmd = 0;

	if (cmd->opcode == MMC_GO_IDLE_STATE) {
		val = SDHI_READ16(host->base, SDHI_CLK_CTRL);
		val |= 0x0100;
		val &= ~0x0200;
		SDHI_WRITE16(host->base, SDHI_CLK_CTRL, val);
		mdelay(10);
		SDHI_WRITE16(host->base, SDHI_CLK_CTRL, val);
	}

	sdhi_clock_start(host);

	/* Waiting for SD Bus busy to be cleared */
	while (!(SDHI_READ16(host->base, SDHI_INFO2) & 0x2000));

	if (cmd->debug)
		printf("CMD%d\n", cmd->opcode);

	mode = cmd->opcode;
	if (acmd)
		mode |= (1 << 6);

	if (cmd->opcode == 55)
		acmd = 1;
	else
		acmd = 0;

	if (cmd->data) {
		mode |= (1 << 11);
		if (cmd->data->flags & MMC_DATA_READ)
			mode |= (1 << 12);

		if (cmd->data->blocks > 1) {
			mode |= (1 << 13);
			SDHI_WRITE16(host->base, SDHI_STOP, 0x100);

			SDHI_WRITE16(host->base, SDHI_SECCNT, cmd->data->blocks);
			SDHI_WRITE16(host->base, SDHI_SIZE, 512);
		} else {
			SDHI_WRITE16(host->base, SDHI_SIZE, cmd->data->blksz);
		}
	}

	/* status clear */
	SDHI_WRITE16(host->base, SDHI_INFO1, 0x0000);
	SDHI_WRITE16(host->base, SDHI_INFO2, 0x0000);

	SDHI_WRITE16(host->base, SDHI_ARG0, cmd->arg & 0xffff);
	SDHI_WRITE16(host->base, SDHI_ARG1, (cmd->arg >> 16) & 0xffff);
	if (cmd->debug)
		printf("XFERTYP: %p\n", mode);

	SDHI_WRITE16(host->base, SDHI_CMD, mode);


	i=1000;
	if (cmd->opcode != 12) {
		for (i=1000000; i; i--) {
			stat = SDHI_READ16(host->base, SDHI_INFO1);
			if (stat & 0x1)
				break;
			udelay(1);
		}
	}
	SDHI_WRITE16(host->base, SDHI_INFO1, 0xfffe);

	stat2 = SDHI_READ16(host->base, SDHI_INFO2);
	if ((i <= 0) || (stat2 & 0x7f)) {
		if (cmd->opcode != SD_SEND_IF_COND)
			printf("CMD:%d timedout. (status = %p)\n",
				cmd->opcode, stat2, i);
		return -1;
	}

	if (cmd->flags & MMC_RSP_136) {
		u32 rsp[8];

		for (i=0; i<8; i++)
			rsp[i] = (u32)SDHI_READ16(host->base, (SDHI_RSP0 + 2 * i));

		cmd->resp[3] = (((rsp[1] << 24) & 0xff000000) |
				((rsp[0] << 8) & 0xff00));
		cmd->resp[2] = (((rsp[3] << 24) & 0xff000000) |
				(rsp[2] << 8) | ((rsp[1] >> 8) & 0x00ff));
		cmd->resp[1] = (((rsp[5] << 24) & 0xff000000) |
				(rsp[4] << 8) | ((rsp[3] >> 8) & 0x00ff));
		cmd->resp[0] = (((rsp[7] << 24) & 0xff000000) |
				(rsp[6] << 8) | ((rsp[5] >> 8) & 0x00ff));
	} else {
		cmd->resp[0] = ((SDHI_READ16(host->base, SDHI_RSP1) << 16) |
				SDHI_READ16(host->base, SDHI_RSP0));
	}

	if (cmd->debug)
		printf("RESP: %p, %p, %p, %p\n",
			cmd->resp[0], cmd->resp[1],
			cmd->resp[2], cmd->resp[3]);

	if (cmd->data) {
		sdhi_pio_transfer_data(host, cmd);

		if (cmd->data->stop) {
			int ret;
			ret = sdhi_do_cmd(host, cmd->data->stop);
			if (ret)
				return ret;
		}
	}

	return 0;
}

static int sdhi_get_card_detect(struct mmc_host *host)
{
	u16 val = 0;

	val = SDHI_READ16(host->base, SDHI_INFO1);
	if (val & 0x20)
		return 1; /* Card Detected */

	return 0;
}

static int sdhi_set_bus_width(struct mmc_host *host, u32 width)
{
	u32 val;

	val = SDHI_READ16(host->base, SDHI_OPTION);

	switch (width) {
	case 4:
		val &= ~(1 << 15);
		break;
	default:
		val |= (1 << 15);
		break;
	}
	SDHI_WRITE16(host->base, SDHI_OPTION, val);

	return 0;
}

static void sdhi_reset(void)
{
	mdelay(1);
}

static void sdhi_set_cd_detect_time(struct mmc_host *host)
{
	u16 val;
	int port;

	while (!(SDHI_READ16(host->base, SDHI_INFO2) & 0x2000));

	val = SDHI_READ16(host->base, SDHI_OPTION);
	val &= ~0xf;

	/* about 1.6msec */
	port = get_sdhi_port(host->base);
	if (port == 0)
		val |= 0x4;
	else
		val |= 0x3;

	SDHI_WRITE16(host->base, SDHI_OPTION, val);
}

void sdhi_exit(struct mmc_host *host)
{
	u32 reg;

	sdhi_clock_stop(host);
	sdhi_clock_stop(host);

	/* GP6[all] -> GPIO */
	reg = read32(PFC_GPSR6);
	reg |=  0x00000000;
	write32(PFC_PMMR, ~reg);
	write32(PFC_GPSR6, reg);
}

static struct mmc_host_ops rcm2_sdhi_ops = {
	.get_cd = sdhi_get_card_detect,
	.do_cmd = sdhi_do_cmd,
	.set_clock = sdhi_set_clock,
	.set_bus_width = sdhi_set_bus_width,
};

struct mmc_host rcm2_sdhi = {
	.ops = &rcm2_sdhi_ops,
};

static int rcm2_mmcsd_probe(struct mmcsd_device *dev, int port)
{
	dev->host = &rcm2_sdhi;

	switch (port) {
	case 0:
		dev->host->base = SDHI0_BASE_ADDR;
		break;
	case 1:
		dev->host->base = SDHI1_BASE_ADDR;
		break;
	default:
		return -1;
	}

	return 0;
}

struct mmcsd_device rcm2_mmcsd;

void sdhi_setup(int port)
{
	rcm2_mmcsd_probe(&rcm2_mmcsd, port);

	port = get_sdhi_port(rcm2_sdhi.base);

	sdhi_reset();

	sdhi_pinmux(port);

	sdhi_cpg_enable(port);

	SDHI_WRITE16(rcm2_sdhi.base, SDHI_SOFT_RST, 0x1);

	sdhi_set_cd_detect_time(&rcm2_sdhi);
	mdelay(5);

	sdhi_set_clock(&rcm2_sdhi, 0);
	sdhi_set_bus_width(&rcm2_sdhi, 1);
}
